//
// Simple volumetric raycasting fragment shader - building isosurface
//

uniform sampler3D modelMap;
uniform vec3	light;
uniform vec3	eye;
uniform float	alpha;

varying	vec3	texCoord;
varying	vec3	l;
varying	vec3	v;

void main ()
{
	const	float	eps = 0.005;

	vec3	l2    = normalize ( l );
	vec3	v2    = normalize ( v );
	float	delta = 1.0 / 300.0;
	vec3	tex   = gl_TexCoord [0].xyz;
	vec3	t     = tex - 0.5*delta*v2;
	vec4	c     = texture3D ( modelMap, t );
	float	i;
	
	for ( i = 0; i < 10000; i++ )
	{
		vec4	clr = texture3D ( modelMap, tex );
		
		if ( (c.a - alpha) * (clr.a - alpha) < eps )
		{
			vec3	normal  = (clr.xyz + c.xyz) - vec3 ( 1.0, 1.0, 1.0 );
			float	len     = length ( normal );
			float	diffuse = 0.2 + abs ( dot ( l2, normal ) / len ) * 0.7;
			
			gl_FragColor = vec4 ( vec3 ( diffuse ), 1.0 );
			
			i = 66000;
		}
		
		t    = tex;
		c    = clr;
		tex -= delta * v2;
		
		if ( any ( lessThan ( tex, vec3 ( 0.0 ) ) ) || any ( greaterThan ( tex, vec3 ( 1.0 ) ) ) )
			i = 66000;
	}
}
