//
// Example of rendering of volumetrics data
//

#include	"libExt.h"
#include	"libTexture3D.h"
#include    <glut.h>
#include    <stdio.h>
#include    <stdlib.h>
#include	"Plane.h"
#include	"Matrix3D.h"
#include	"GlslProgram.h"
#include	"TypeDefs.h"

unsigned	modelMap = 0;
Vector3D	cubeVertex [8];
Vector3D	rot ( 0, 0, 0 );
int			mouseOldX = 0;
int			mouseOldY = 0;
float		alpha     = 0.4;
Vector3D 	eye   ( 0,   0, 2.5 );
Vector3D 	light ( 2.0, 0, 2.0 );
GlslProgram	program;

bool    loadVolume ( const char * fileName )
{
	struct
	{
		uint32	width;
		uint32	height;
		uint32	depth;
		uint32	numComponents;
	} hdr;
	
    FILE * fp = fopen ( fileName, "rb" );

    if ( fp == NULL )
        return false;
                                                // write file signature
    if ( fread ( &hdr, 1, sizeof ( hdr ), fp ) != sizeof ( hdr ) )
		return false;
		
	unsigned	size = hdr.width * hdr.height * hdr.depth * hdr.numComponents;
	byte      * data = (byte *) malloc ( size );
	
                                                // write texture data
	if ( fread  ( data, 1, size, fp ) != size )
	{
		fclose ( fp );
		
		return false;
	}
	
	fclose ( fp );

	int	format = GL_ALPHA;
	
	if ( hdr.numComponents == 3 )
		format = GL_RGB;
	else
	if ( hdr.numComponents == 4 )
		format = GL_RGBA;
		
	glGenTextures ( 1, &modelMap );
    glBindTexture ( GL_TEXTURE_3D, modelMap );
	glEnable      ( GL_TEXTURE_3D );

                                                    // use SGIS_generate_mipmap extensions
    glTexParameteri ( GL_TEXTURE_3D_EXT, GL_GENERATE_MIPMAP_SGIS, GL_TRUE );
    glTexParameteri ( GL_TEXTURE_3D_EXT, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
	
    glTexImage3DEXT ( GL_TEXTURE_3D, 0, hdr.numComponents, hdr.width, hdr.height, hdr.depth, 0,
                      format, GL_UNSIGNED_BYTE, data );

	glTexParameteri ( GL_TEXTURE_3D, GL_GENERATE_MIPMAP_SGIS, GL_TRUE );
    glTexParameteri ( GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER,   GL_LINEAR );
    glTexParameteri ( GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER,   GL_LINEAR_MIPMAP_LINEAR );
    glTexParameteri ( GL_TEXTURE_3D, GL_TEXTURE_WRAP_S,       GL_CLAMP  );
    glTexParameteri ( GL_TEXTURE_3D, GL_TEXTURE_WRAP_T,       GL_CLAMP  );
    glTexParameteri ( GL_TEXTURE_3D, GL_TEXTURE_WRAP_R,       GL_CLAMP  );
	
	free ( data );
	
    return modelMap != 0;
}

void	initCube ()
{
	cubeVertex [0] = Vector3D ( -0.5f, -0.5f, -0.5f );
	cubeVertex [1] = Vector3D (  0.5f, -0.5f, -0.5f );
	cubeVertex [2] = Vector3D ( -0.5f,  0.5f, -0.5f );
	cubeVertex [3] = Vector3D (  0.5f,  0.5f, -0.5f );
	cubeVertex [4] = Vector3D ( -0.5f, -0.5f,  0.5f );
	cubeVertex [5] = Vector3D (  0.5f, -0.5f,  0.5f );
	cubeVertex [6] = Vector3D ( -0.5f,  0.5f,  0.5f );
	cubeVertex [7] = Vector3D (  0.5f,  0.5f,  0.5f );
}

void	renderVolume ( Vector3D * vertices, const Matrix3D& m )
{
	glEnable ( GL_DEPTH_TEST );
										// setup render state
	program.bind     ();
								
	program.setUniformVector ( "eye",   eye   );
	program.setUniformVector ( "light", light );
	program.setUniformFloat  ( "alpha", alpha );

	glActiveTextureARB ( GL_TEXTURE0_ARB );
	glBindTexture      ( GL_TEXTURE_3D, modelMap );
	
										// render cube faces
	glutSolidCube ( 1.0 );
	
	program.unbind ();
}

void display ()
{
    glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	
	Matrix3D	r;
	
	glMatrixMode ( GL_MODELVIEW );
	glPushMatrix ();
	glRotatef    ( rot.y, 0, 1, 0 );
	glRotatef    ( rot.z, 0, 0, 1 );
	renderVolume ( cubeVertex, r );
	glPopMatrix  ();

    glutSwapBuffers ();
}

void init ()
{
    glClearColor ( 0.0, 0.0, 0.0, 1.0 );
    glEnable     ( GL_DEPTH_TEST );
//    glEnable     ( GL_TEXTURE_2D );
    glDepthFunc  ( GL_LEQUAL     );

    glHint ( GL_POLYGON_SMOOTH_HINT,         GL_NICEST );
    glHint ( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
}

void reshape ( int w, int h )
{
   glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
   glMatrixMode   ( GL_PROJECTION );
   glLoadIdentity ();
   gluPerspective ( 40.0, (GLfloat)w/(GLfloat)h, 0.2, 10.0 );
   glMatrixMode   ( GL_MODELVIEW );
   glLoadIdentity ();
   gluLookAt      ( eye.x, eye.y, eye.z,    	// eye
                    0, 0, 0,        			// center
                    0, 1, 0 );      			// up
}

void motion ( int x, int y )
{
    rot.z -= ((mouseOldY - y) * 180.0f) / 300.0f;
    rot.y -= ((mouseOldX - x) * 180.0f) / 300.0f;
    rot.x  = 0;

    if ( rot.z > 360 )
		rot.z -= 360;

	if ( rot.z < -360 )
		rot.z += 360;

    if ( rot.y > 360 )
		rot.y -= 360;

	if ( rot.y < -360 )
        rot.y += 360;

    mouseOldX = x;
    mouseOldY = y;

	glutPostRedisplay ();
}

void mouse ( int button, int state, int x, int y )
{
    if ( state == GLUT_DOWN )
    {
        mouseOldX = x;
        mouseOldY = y;
	}
}

void key ( unsigned char key, int x, int y )
{
    if ( key == 27 || key == 'q' || key == 'Q' )        // quit requested
        exit ( 0 );
		
	if ( key == '+' )
		alpha += 0.0025f;
	else
	if ( key == '-' )
		alpha -= 0.0025f;
		
	printf ( "Alpha: %f\n", alpha );
	
	glutPostRedisplay ();
}

void    specialKey ( int key, int x, int y )
{
    glutPostRedisplay ();
}

int main ( int argc, char * argv [] )
{
                                // initialize glut
    glutInit            ( &argc, argv );
    glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH );
    glutInitWindowSize  ( 512, 512 );

                                // create window
    glutCreateWindow ( "OpenGL volume raycasting demo" );

                                // register handlers
    glutDisplayFunc  ( display    );
    glutReshapeFunc  ( reshape    );
    glutKeyboardFunc ( key        );
    glutSpecialFunc  ( specialKey );
	glutMouseFunc    ( mouse      );
	glutMotionFunc   ( motion     );

    init           ();
    initExtensions ();

	if ( !GlslProgram :: isSupported () )
	{
		printf ( "GLSL not supported.\n" );

		return 1;
	}

    if ( !program.loadShaders ( "raycast.vsh", "raycast.fsh" ) )
    {
        printf ( "Error loading shaders:\n%s\n", program.getLog ().c_str () );

        return 3;
    }

									// load 3D texture from file
	if ( !loadVolume ( "volume-rgba.vol" ) )
	{
		printf ( "Error loading model texture\n" );
		
		exit ( 1 );
	}

	program.bind         ();
	program.setTexture   ( "modelMap",  0 );
	program.unbind       ();

	initCube ();

	printf ( "Data and shaders loaded.\n" );

    glutMainLoop ();

    return 0;
}
